@@ -73,7 +73,7 @@ CK_RV
 __pkcs11h_certificate_doPrivateOperation (
 	IN const pkcs11h_certificate_t certificate,
 	IN const enum __pkcs11h_private_op_e op,
-	IN const CK_MECHANISM_TYPE mech_type,
+	IN const CK_MECHANISM * const mech,
 	IN const unsigned char * const source,
 	IN const size_t source_size,
 	OUT unsigned char * const target,
@@ -476,7 +476,9 @@ __pkcs11h_certificate_getKeyAttributes (
 			{CKA_SIGN, NULL, 0},
 			{CKA_SIGN_RECOVER, NULL, 0},
 			{CKA_DECRYPT, NULL, 0},
-			{CKA_UNWRAP, NULL, 0}
+			{CKA_UNWRAP, NULL, 0},
+			{CKA_ALWAYS_AUTHENTICATE, NULL, 0},
+			{CKA_LABEL, NULL, 0}
 		};
 
 		/*
@@ -502,6 +504,7 @@ __pkcs11h_certificate_getKeyAttributes (
 			CK_BBOOL *key_attrs_sign_recover;
 			CK_BBOOL *key_attrs_decrypt;
 			CK_BBOOL *key_attrs_unwrap;
+			CK_BBOOL *key_attrs_always_authenticate;
 
 			if (
 				(rv = _pkcs11h_session_getObjectAttributes (
@@ -518,6 +521,7 @@ __pkcs11h_certificate_getKeyAttributes (
 			key_attrs_sign_recover = (CK_BBOOL *)key_attrs[1].pValue;
 			key_attrs_decrypt = (CK_BBOOL *)key_attrs[2].pValue;
 			key_attrs_unwrap = (CK_BBOOL *)key_attrs[3].pValue;
+			key_attrs_always_authenticate = (CK_BBOOL *)key_attrs[4].pValue;
 
 			if (key_attrs_sign != NULL && *key_attrs_sign != CK_FALSE) {
 				certificate->mask_private_mode |= PKCS11H_PRIVATEMODE_MASK_SIGN;
@@ -536,10 +540,34 @@ __pkcs11h_certificate_getKeyAttributes (
 				goto retry;
 			}
 
+			if (key_attrs_always_authenticate != NULL) {
+				certificate->always_authenticate = *key_attrs_always_authenticate != 0;
+			}
+
+			if (strlen(certificate->id->displayName) == 0) {
+				if (key_attrs[5].pValue == NULL) {
+					strcpy(certificate->id->displayName, "Certificate");
+				}
+				else {
+					size_t len =  (
+						key_attrs[5].ulValueLen < sizeof(certificate->id->displayName) - 1 ?
+						key_attrs[5].ulValueLen :
+						sizeof(certificate->id->displayName) - 1
+					);
+					memmove(
+						certificate->id->displayName,
+						key_attrs[5].pValue,
+						len
+					);
+					certificate->id->displayName[len] = '\0';
+				}
+			}
+
 			_PKCS11H_DEBUG (
 				PKCS11H_LOG_DEBUG1,
-				"PKCS#11: Key attributes loaded (%08x)",
-				certificate->mask_private_mode
+				"PKCS#11: Key attribute=%08x always_authenticated=%d",
+				certificate->mask_private_mode,
+				certificate->always_authenticate
 			);
 
 			op_succeed = TRUE;
@@ -777,7 +805,7 @@ CK_RV
 __pkcs11h_certificate_doPrivateOperation (
 	IN const pkcs11h_certificate_t certificate,
 	IN const enum __pkcs11h_private_op_e op,
-	IN const CK_MECHANISM_TYPE mech_type,
+	IN const CK_MECHANISM * const mech,
 	IN const unsigned char * const source,
 	IN const size_t source_size,
 	OUT unsigned char * const target,
@@ -786,9 +814,6 @@ __pkcs11h_certificate_doPrivateOperation (
 #if defined(ENABLE_PKCS11H_THREADING)
 	PKCS11H_BOOL mutex_locked = FALSE;
 #endif
-	CK_MECHANISM mech = {
-		mech_type, NULL, 0
-	};
 
 /*	CK_BBOOL wrap_attrs_false = CK_FALSE; */
 	CK_BBOOL wrap_attrs_true = CK_TRUE;
@@ -812,6 +837,7 @@ __pkcs11h_certificate_doPrivateOperation (
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 	_PKCS11H_ASSERT (certificate!=NULL);
+	_PKCS11H_ASSERT (mech!=NULL);
 	_PKCS11H_ASSERT (source!=NULL);
 	/*_PKCS11H_ASSERT (target); NOT NEEDED*/
 	_PKCS11H_ASSERT (p_target_size!=NULL);
@@ -821,7 +847,7 @@ __pkcs11h_certificate_doPrivateOperation (
 		"PKCS#11: __pkcs11h_certificate_doPrivateOperation entry certificate=%p, op=%d, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
 		(void *)certificate,
 		op,
-		mech_type,
+		mech->mechanism,
 		source,
 		source_size,
 		target,
@@ -851,28 +877,28 @@ __pkcs11h_certificate_doPrivateOperation (
 				case __pkcs11h_private_op_sign:
 					rv = certificate->session->provider->f->C_SignInit (
 						certificate->session->session_handle,
-						&mech,
+						(CK_MECHANISM*)mech,
 						certificate->key_handle
 					);
 				break;
 				case __pkcs11h_private_op_sign_recover:
 					rv = certificate->session->provider->f->C_SignRecoverInit (
 						certificate->session->session_handle,
-						&mech,
+						(CK_MECHANISM*)mech,
 						certificate->key_handle
 					);
 				break;
 				case __pkcs11h_private_op_decrypt:
 					rv = certificate->session->provider->f->C_DecryptInit (
 						certificate->session->session_handle,
-						&mech,
+						(CK_MECHANISM*)mech,
 						certificate->key_handle
 					);
 				break;
 				case __pkcs11h_private_op_unwrap:
 					rv = certificate->session->provider->f->C_UnwrapKey (
 						certificate->session->session_handle,
-						&mech,
+						(CK_MECHANISM*)mech,
 						certificate->key_handle,
 						(CK_BYTE_PTR)source,
 						source_size,
@@ -895,6 +921,29 @@ __pkcs11h_certificate_doPrivateOperation (
 			if (rv != CKR_OK) {
 				goto retry;
 			}
+
+			if (certificate->always_authenticate) {
+				_PKCS11H_DEBUG (
+					PKCS11H_LOG_DEBUG2,
+					"PKCS#11: CKU_CONTEXT_SPECIFIC login"
+				);
+				if (
+					(rv = _pkcs11h_session_login_context (
+						certificate->session,
+						CKU_CONTEXT_SPECIFIC,
+						certificate->id->displayName,
+						certificate->user_data,
+						(certificate->mask_prompt & PKCS11H_PROMPT_MASK_ALLOW_PIN_PROMPT)
+					)) != CKR_OK
+				) {
+					_PKCS11H_DEBUG (
+						PKCS11H_LOG_DEBUG2,
+						"PKCS#11: CKU_CONTEXT_SPECIFIC login rv=%ld",
+						rv
+					);
+					goto retry;
+				}
+			}
 		}
 
 		/*
@@ -1304,21 +1353,42 @@ pkcs11h_certificate_sign (
 	IN const size_t source_size,
 	OUT unsigned char * const target,
 	IN OUT size_t * const p_target_size
+) {
+	CK_MECHANISM mech = {mech_type, NULL, 0};
+	return pkcs11h_certificate_sign_ex (
+		certificate,
+		&mech,
+		source,
+		source_size,
+		target,
+		p_target_size
+	);
+}
+
+CK_RV
+pkcs11h_certificate_sign_ex (
+	IN const pkcs11h_certificate_t certificate,
+	IN const CK_MECHANISM * const mech,
+	IN const unsigned char * const source,
+	IN const size_t source_size,
+	OUT unsigned char * const target,
+	IN OUT size_t * const p_target_size
 ) {
 	CK_RV rv = CKR_FUNCTION_FAILED;
 
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 	_PKCS11H_ASSERT (certificate!=NULL);
+	_PKCS11H_ASSERT (mech!=NULL);
 	_PKCS11H_ASSERT (source!=NULL);
 	/*_PKCS11H_ASSERT (target); NOT NEEDED*/
 	_PKCS11H_ASSERT (p_target_size!=NULL);
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_certificate_sign entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
+		"PKCS#11: pkcs11h_certificate_sign_ex entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
 		(void *)certificate,
-		mech_type,
+		mech->mechanism,
 		source,
 		source_size,
 		target,
@@ -1333,7 +1403,7 @@ pkcs11h_certificate_sign (
 		(rv = __pkcs11h_certificate_doPrivateOperation (
 			certificate,
 			__pkcs11h_private_op_sign,
-			mech_type,
+			mech,
 			source,
 			source_size,
 			target,
@@ -1366,21 +1436,42 @@ pkcs11h_certificate_signRecover (
 	IN const size_t source_size,
 	OUT unsigned char * const target,
 	IN OUT size_t * const p_target_size
+) {
+	CK_MECHANISM mech = {mech_type, NULL, 0};
+	return pkcs11h_certificate_signRecover_ex (
+		certificate,
+		&mech,
+		source,
+		source_size,
+		target,
+		p_target_size
+	);
+}
+
+CK_RV
+pkcs11h_certificate_signRecover_ex (
+	IN const pkcs11h_certificate_t certificate,
+	IN const CK_MECHANISM * const mech,
+	IN const unsigned char * const source,
+	IN const size_t source_size,
+	OUT unsigned char * const target,
+	IN OUT size_t * const p_target_size
 ) {
 	CK_RV rv = CKR_FUNCTION_FAILED;
 
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 	_PKCS11H_ASSERT (certificate!=NULL);
+	_PKCS11H_ASSERT (mech!=NULL);
 	_PKCS11H_ASSERT (source!=NULL);
 	/*_PKCS11H_ASSERT (target); NOT NEEDED*/
 	_PKCS11H_ASSERT (p_target_size!=NULL);
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_certificate_signRecover entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
+		"PKCS#11: pkcs11h_certificate_signRecover_ex entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
 		(void *)certificate,
-		mech_type,
+		mech->mechanism,
 		source,
 		source_size,
 		target,
@@ -1395,7 +1486,7 @@ pkcs11h_certificate_signRecover (
 		(rv = __pkcs11h_certificate_doPrivateOperation (
 			certificate,
 			__pkcs11h_private_op_sign_recover,
-			mech_type,
+			mech,
 			source,
 			source_size,
 			target,
@@ -1428,21 +1519,42 @@ pkcs11h_certificate_decrypt (
 	IN const size_t source_size,
 	OUT unsigned char * const target,
 	IN OUT size_t * const p_target_size
+) {
+	CK_MECHANISM mech = {mech_type, NULL, 0};
+	return pkcs11h_certificate_decrypt_ex (
+		certificate,
+		&mech,
+		source,
+		source_size,
+		target,
+		p_target_size
+	);
+}
+
+CK_RV
+pkcs11h_certificate_decrypt_ex (
+	IN const pkcs11h_certificate_t certificate,
+	IN const CK_MECHANISM * const mech,
+	IN const unsigned char * const source,
+	IN const size_t source_size,
+	OUT unsigned char * const target,
+	IN OUT size_t * const p_target_size
 ) {
 	CK_RV rv = CKR_FUNCTION_FAILED;
 
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 	_PKCS11H_ASSERT (certificate!=NULL);
+	_PKCS11H_ASSERT (mech!=NULL);
 	_PKCS11H_ASSERT (source!=NULL);
 	/*_PKCS11H_ASSERT (target); NOT NEEDED*/
 	_PKCS11H_ASSERT (p_target_size!=NULL);
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_certificate_decrypt entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
+		"PKCS#11: pkcs11h_certificate_decrypt_ex entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
 		(void *)certificate,
-		mech_type,
+		mech->mechanism,
 		source,
 		source_size,
 		target,
@@ -1457,7 +1569,7 @@ pkcs11h_certificate_decrypt (
 		(rv = __pkcs11h_certificate_doPrivateOperation (
 			certificate,
 			__pkcs11h_private_op_decrypt,
-			mech_type,
+			mech,
 			source,
 			source_size,
 			target,
@@ -1490,21 +1602,42 @@ pkcs11h_certificate_unwrap (
 	IN const size_t source_size,
 	OUT unsigned char * const target,
 	IN OUT size_t * const p_target_size
+) {
+	CK_MECHANISM mech = {mech_type, NULL, 0};
+	return pkcs11h_certificate_unwrap_ex (
+		certificate,
+		&mech,
+		source,
+		source_size,
+		target,
+		p_target_size
+	);
+}
+
+CK_RV
+pkcs11h_certificate_unwrap_ex (
+	IN const pkcs11h_certificate_t certificate,
+	IN const CK_MECHANISM * const mech,
+	IN const unsigned char * const source,
+	IN const size_t source_size,
+	OUT unsigned char * const target,
+	IN OUT size_t * const p_target_size
 ) {
 	CK_RV rv = CKR_FUNCTION_FAILED;
 
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 	_PKCS11H_ASSERT (certificate!=NULL);
+	_PKCS11H_ASSERT (mech!=NULL);
 	_PKCS11H_ASSERT (source!=NULL);
 	/*_PKCS11H_ASSERT (target); NOT NEEDED*/
 	_PKCS11H_ASSERT (p_target_size!=NULL);
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_certificate_unwrap entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
+		"PKCS#11: pkcs11h_certificate_unwrap_ex entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
 		(void *)certificate,
-		mech_type,
+		mech->mechanism,
 		source,
 		source_size,
 		target,
@@ -1519,7 +1652,7 @@ pkcs11h_certificate_unwrap (
 		(rv = __pkcs11h_certificate_doPrivateOperation (
 			certificate,
 			__pkcs11h_private_op_unwrap,
-			mech_type,
+			mech,
 			source,
 			source_size,
 			target,
@@ -1552,6 +1685,26 @@ pkcs11h_certificate_signAny (
 	IN const size_t source_size,
 	OUT unsigned char * const target,
 	IN OUT size_t * const p_target_size
+) {
+	CK_MECHANISM mech = {mech_type, NULL, 0};
+	return pkcs11h_certificate_signAny_ex (
+		certificate,
+		&mech,
+		source,
+		source_size,
+		target,
+		p_target_size
+	);
+}
+
+CK_RV
+pkcs11h_certificate_signAny_ex (
+	IN const pkcs11h_certificate_t certificate,
+	IN const CK_MECHANISM *mech,
+	IN const unsigned char * const source,
+	IN const size_t source_size,
+	OUT unsigned char * const target,
+	IN OUT size_t * const p_target_size
 ) {
 	CK_RV rv = CKR_FUNCTION_FAILED;
 	PKCS11H_BOOL acked = FALSE;
@@ -1559,15 +1712,16 @@ pkcs11h_certificate_signAny (
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 	_PKCS11H_ASSERT (certificate!=NULL);
+	_PKCS11H_ASSERT (mech!=NULL);
 	_PKCS11H_ASSERT (source!=NULL);
 	/*_PKCS11H_ASSERT (target); NOT NEEDED*/
 	_PKCS11H_ASSERT (p_target_size!=NULL);
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_certificate_signAny entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
+		"PKCS#11: pkcs11h_certificate_signAny_ex entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
 		(void *)certificate,
-		mech_type,
+		mech->mechanism,
 		source,
 		source_size,
 		target,
@@ -1590,9 +1744,9 @@ pkcs11h_certificate_signAny (
 		(certificate->mask_private_mode & PKCS11H_PRIVATEMODE_MASK_SIGN) != 0
 	) {
 		switch (
-			(rv = pkcs11h_certificate_sign (
+			(rv = pkcs11h_certificate_sign_ex (
 				certificate,
-				mech_type,
+				mech,
 				source,
 				source_size,
 				target,
@@ -1617,9 +1771,9 @@ pkcs11h_certificate_signAny (
 		(certificate->mask_private_mode & PKCS11H_PRIVATEMODE_MASK_RECOVER) != 0
 	) {
 		switch (
-			(rv = pkcs11h_certificate_signRecover (
+			(rv = pkcs11h_certificate_signRecover_ex (
 				certificate,
-				mech_type,
+				mech,
 				source,
 				source_size,
 				target,
@@ -1667,6 +1821,26 @@ pkcs11h_certificate_decryptAny (
 	IN const size_t source_size,
 	OUT unsigned char * const target,
 	IN OUT size_t * const p_target_size
+) {
+	CK_MECHANISM mech = {mech_type, NULL, 0};
+	return pkcs11h_certificate_decryptAny_ex (
+		certificate,
+		&mech,
+		source,
+		source_size,
+		target,
+		p_target_size
+	);
+}
+
+CK_RV
+pkcs11h_certificate_decryptAny_ex (
+	IN const pkcs11h_certificate_t certificate,
+	IN const CK_MECHANISM * const mech,
+	IN const unsigned char * const source,
+	IN const size_t source_size,
+	OUT unsigned char * const target,
+	IN OUT size_t * const p_target_size
 ) {
 	CK_RV rv = CKR_FUNCTION_FAILED;
 	PKCS11H_BOOL acked = FALSE;
@@ -1674,15 +1848,16 @@ pkcs11h_certificate_decryptAny (
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
 	_PKCS11H_ASSERT (certificate!=NULL);
+	_PKCS11H_ASSERT (mech!=NULL);
 	_PKCS11H_ASSERT (source!=NULL);
 	/*_PKCS11H_ASSERT (target); NOT NEEDED*/
 	_PKCS11H_ASSERT (p_target_size!=NULL);
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_certificate_decryptAny entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
+		"PKCS#11: pkcs11h_certificate_decryptAny_ex entry certificate=%p, mech_type=%ld, source=%p, source_size="P_Z", target=%p, *p_target_size="P_Z"",
 		(void *)certificate,
-		mech_type,
+		mech->mechanism,
 		source,
 		source_size,
 		target,
@@ -1704,9 +1879,9 @@ pkcs11h_certificate_decryptAny (
 		(certificate->mask_private_mode & PKCS11H_PRIVATEMODE_MASK_DECRYPT) != 0
 	) {
 		switch (
-			pkcs11h_certificate_decrypt (
+			pkcs11h_certificate_decrypt_ex (
 				certificate,
-				mech_type,
+				mech,
 				source,
 				source_size,
 				target,
@@ -1731,9 +1906,9 @@ pkcs11h_certificate_decryptAny (
 		(certificate->mask_private_mode & PKCS11H_PRIVATEMODE_MASK_UNWRAP) != 0
 	) {
 		switch (
-			pkcs11h_certificate_unwrap (
+			pkcs11h_certificate_unwrap_ex (
 				certificate,
-				mech_type,
+				mech,
 				source,
 				source_size,
 				target,
