@@ -117,6 +117,9 @@ CK_RV
 __pkcs11h_forkFixup ();
 #endif
 
+static
+_pkcs11h_provider_t
+__pkcs11h_get_pkcs11_provider(const char * const reference);
 
 /*==========================================
  * Data
@@ -125,6 +128,19 @@ __pkcs11h_forkFixup ();
 _pkcs11h_data_t _g_pkcs11h_data = NULL;
 unsigned int _g_pkcs11h_loglevel = PKCS11H_LOG_INFO;
 
+static const char * __pkcs11h_provider_preperty_names[] = {
+	"location",
+	"allow_protected_auth",
+	"mask_private_mode",
+	"slot_event_method",
+	"slot_poll_interval",
+	"cert_is_private",
+	"init_args",
+	"provider_destruct_hook",
+	"provider_destruct_hook_data",
+	NULL
+};
+
 /*======================================================================*
  * PUBLIC INTERFACE
  *======================================================================*/
@@ -288,6 +304,13 @@ pkcs11h_initialize (void) {
 
 	pkcs11h_terminate ();
 
+	_PKCS11H_ASSERT (
+		(
+			sizeof(__pkcs11h_provider_preperty_names) /
+			sizeof(*__pkcs11h_provider_preperty_names)
+		) == _PKCS11H_PROVIDER_PROPERTY_LAST + 1
+	);
+
 	if ((rv = _pkcs11h_mem_malloc ((void*)&data, sizeof (struct _pkcs11h_data_s))) != CKR_OK) {
 		goto cleanup;
 	}
@@ -511,32 +534,240 @@ pkcs11h_terminate (void) {
 	return CKR_OK;
 }
 
+static
+CK_RV
+__pkcs11h_propertyAddress(
+	IN const unsigned property,
+	OUT void ** value,
+	OUT size_t * value_size
+) {
+	CK_RV rv = CKR_FUNCTION_FAILED;
+
+	switch (property) {
+		default:
+			_PKCS11H_DEBUG (
+				PKCS11H_LOG_ERROR,
+				"PKCS#11: Trying to lookup library provider property '%d'",
+				property
+			);
+			rv = CKR_ATTRIBUTE_TYPE_INVALID;
+			goto cleanup;
+		case PKCS11H_PROPERTY_LOG_LEVEL:
+			*value = &_g_pkcs11h_loglevel;
+			*value_size = sizeof(_g_pkcs11h_loglevel);
+		break;
+		case PKCS11H_PROPERTY_FORK_MODE:
+#if !defined(_WIN32)
+			*value = &_g_pkcs11h_data->safefork;
+			*value_size = sizeof(_g_pkcs11h_data->safefork);
+#else
+			rv = CKR_FUNCTION_NOT_SUPPORTED;
+			goto cleanup;
+#endif
+		break;
+		case PKCS11H_PROPERTY_LOG_HOOK:
+			*value = &_g_pkcs11h_data->hooks.log;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.log);
+		break;
+		case PKCS11H_PROPERTY_LOG_HOOK_DATA:
+			*value = &_g_pkcs11h_data->hooks.log_data;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.log_data);
+		break;
+		case PKCS11H_PROPERTY_SLOT_EVENT_HOOK:
+#if defined(ENABLE_PKCS11H_SLOTEVENT)
+			*value = &_g_pkcs11h_data->hooks.slotevent;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.slotevent);
+#else
+			rv = CKR_FUNCTION_NOT_SUPPORTED;
+			goto cleanup;
+#endif
+		break;
+		case PKCS11H_PROPERTY_SLOT_EVENT_HOOK_DATA:
+#if defined(ENABLE_PKCS11H_SLOTEVENT)
+			*value = &_g_pkcs11h_data->hooks.slotevent_data;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.slotevent_data);
+#else
+			rv = CKR_FUNCTION_NOT_SUPPORTED;
+			goto cleanup;
+#endif
+		break;
+		case PKCS11H_PROPERTY_TOKEN_PROMPT_HOOK:
+			*value = &_g_pkcs11h_data->hooks.token_prompt;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.token_prompt);
+		break;
+		case PKCS11H_PROPERTY_TOKEN_PROMPT_HOOK_DATA:
+			*value = &_g_pkcs11h_data->hooks.token_prompt_data;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.token_prompt_data);
+		break;
+		case PKCS11H_PROPERTY_PIN_PROMPT_HOOK:
+			*value = &_g_pkcs11h_data->hooks.pin_prompt;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.pin_prompt);
+		break;
+		case PKCS11H_PROPERTY_PIN_PROMPT_HOOK_DATA:
+			*value = &_g_pkcs11h_data->hooks.pin_prompt_data;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.pin_prompt_data);
+		break;
+		case PKCS11H_PROPERTY_KEY_PROMPT_HOOK:
+			*value = &_g_pkcs11h_data->hooks.key_prompt;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.key_prompt);
+		break;
+		case PKCS11H_PROPERTY_KEY_PROMPT_HOOK_DATA:
+			*value = &_g_pkcs11h_data->hooks.key_prompt_data;
+			*value_size = sizeof(_g_pkcs11h_data->hooks.key_prompt_data);
+		break;
+		case PKCS11H_PROPERTY_ALLOW_PROTECTED_AUTHENTICATION:
+			*value = &_g_pkcs11h_data->allow_protected_auth;
+			*value_size = sizeof(_g_pkcs11h_data->allow_protected_auth);
+		break;
+		case PKCS11H_PROPERTY_PIN_CACHE_PERIOD:
+			*value = &_g_pkcs11h_data->pin_cache_period;
+			*value_size = sizeof(_g_pkcs11h_data->pin_cache_period);
+		break;
+		case PKCS11H_PROPERTY_MAX_LOGIN_RETRIES:
+			*value = &_g_pkcs11h_data->max_retries;
+			*value_size = sizeof(_g_pkcs11h_data->max_retries);
+		break;
+	}
+	rv = CKR_OK;
+
+cleanup:
+	return rv;
+}
+
+CK_RV
+pkcs11h_getProperty (
+	IN const unsigned property,
+	OUT void * const value,
+	IN OUT size_t * const value_size
+) {
+	CK_RV rv = CKR_FUNCTION_FAILED;
+	void *source;
+	size_t size;
+
+	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
+	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
+	_PKCS11H_ASSERT (value != NULL);
+	_PKCS11H_ASSERT (value_size != NULL);
+
+	if ((rv = __pkcs11h_propertyAddress(property, &source, &size)) != CKR_OK) {
+		goto cleanup;
+	}
+
+	if (size > *value_size) {
+		rv = CKR_BUFFER_TOO_SMALL;
+		goto cleanup;
+	}
+
+	memcpy(value, source, size);
+	rv = CKR_OK;
+
+cleanup:
+
+	return rv;
+}
+
+CK_RV
+pkcs11h_setProperty (
+	IN const unsigned property,
+	IN const void * value,
+	IN const size_t value_size
+) {
+	CK_RV rv = CKR_FUNCTION_FAILED;
+	void *target;
+	size_t size;
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG2,
+		"PKCS#11: pkcs11h_setProperty entry property='%d', value=%p, value_size=%ld",
+		property,
+		value,
+		value_size
+	);
+
+	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
+	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
+	_PKCS11H_ASSERT (value != NULL);
+
+	if ((rv = __pkcs11h_propertyAddress(property, &target, &size)) != CKR_OK) {
+		goto cleanup;
+	}
+
+	if (size != value_size) {
+		rv = CKR_DATA_LEN_RANGE;
+		goto cleanup;
+	}
+
+	if (value_size == sizeof(int)) {
+		_PKCS11H_DEBUG (
+			PKCS11H_LOG_DEBUG1,
+			"PKCS#11: Setting property %d=0x%x",
+			property,
+			*(int *)value
+		);
+	}
+	else if (value_size == sizeof(long)) {
+		_PKCS11H_DEBUG (
+			PKCS11H_LOG_DEBUG1,
+			"PKCS#11: Setting property %d=0x%lx",
+			property,
+			*(long *)value
+		);
+	}
+	else {
+		_PKCS11H_DEBUG (
+			PKCS11H_LOG_DEBUG1,
+			"PKCS#11: Setting property %d=*size*",
+			property
+		);
+	}
+
+	memcpy(target, value, size);
+	rv = CKR_OK;
+
+	switch (property) {
+		case PKCS11H_PROPERTY_SLOT_EVENT_HOOK:
+			if ((rv = _pkcs11h_slotevent_init ()) != CKR_OK) {
+				goto cleanup;
+			}
+		break;
+	}
+cleanup:
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG2,
+		"PKCS#11: pkcs11h_setProperty return rv=%lu-'%s'",
+		rv,
+		pkcs11h_getMessage (rv)
+	);
+
+	return rv;
+}
+
 void
 pkcs11h_setLogLevel (
 	IN const unsigned flags
 ) {
-	_g_pkcs11h_loglevel = flags;
+	pkcs11h_setProperty(PKCS11H_PROPERTY_LOG_LEVEL, &flags, sizeof(flags));
 }
 
 CK_RV
 pkcs11h_setForkMode (
 	IN const PKCS11H_BOOL safe
 ) {
-#if !defined(_WIN32)
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-
-	_g_pkcs11h_data->safefork = safe;
-#endif
-	return CKR_OK;
+	CK_RV rv;
+	rv = pkcs11h_setProperty(PKCS11H_PROPERTY_FORK_MODE, &safe, sizeof(safe));
+	if (rv == CKR_FUNCTION_NOT_SUPPORTED) {	/* backward compatibility */
+		rv = CKR_OK;
+	}
+	return rv;
 }
 
 unsigned
 pkcs11h_getLogLevel (void) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-
-	return _g_pkcs11h_loglevel;
+	unsigned flags;
+	size_t size = sizeof(flags);
+	pkcs11h_getProperty(PKCS11H_PROPERTY_LOG_LEVEL, &flags, &size);
+	return flags;
 }
 
 CK_RV
@@ -544,13 +775,8 @@ pkcs11h_setLogHook (
 	IN const pkcs11h_hook_log_t hook,
 	IN void * const global_data
 ) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-	_PKCS11H_ASSERT (hook!=NULL);
-
-	_g_pkcs11h_data->hooks.log = hook;
-	_g_pkcs11h_data->hooks.log_data = global_data;
-
+	pkcs11h_setProperty(PKCS11H_PROPERTY_LOG_HOOK_DATA, &global_data, sizeof(global_data));
+	pkcs11h_setProperty(PKCS11H_PROPERTY_LOG_HOOK, &hook, sizeof(hook));
 	return CKR_OK;
 }
 
@@ -559,20 +785,8 @@ pkcs11h_setSlotEventHook (
 	IN const pkcs11h_hook_slotevent_t hook,
 	IN void * const global_data
 ) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-	_PKCS11H_ASSERT (hook!=NULL);
-
-#if defined(ENABLE_PKCS11H_SLOTEVENT)
-	_g_pkcs11h_data->hooks.slotevent = hook;
-	_g_pkcs11h_data->hooks.slotevent_data = global_data;
-
-	return _pkcs11h_slotevent_init ();
-#else
-	(void)global_data;
-
-	return CKR_FUNCTION_NOT_SUPPORTED;
-#endif
+	pkcs11h_setProperty(PKCS11H_PROPERTY_SLOT_EVENT_HOOK_DATA, &global_data, sizeof(global_data));
+	return pkcs11h_setProperty(PKCS11H_PROPERTY_SLOT_EVENT_HOOK, &hook, sizeof(hook));
 }
 
 CK_RV
@@ -580,13 +794,8 @@ pkcs11h_setPINPromptHook (
 	IN const pkcs11h_hook_pin_prompt_t hook,
 	IN void * const global_data
 ) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-	_PKCS11H_ASSERT (hook!=NULL);
-
-	_g_pkcs11h_data->hooks.pin_prompt = hook;
-	_g_pkcs11h_data->hooks.pin_prompt_data = global_data;
-
+	pkcs11h_setProperty(PKCS11H_PROPERTY_PIN_PROMPT_HOOK_DATA, &global_data, sizeof(global_data));
+	pkcs11h_setProperty(PKCS11H_PROPERTY_PIN_PROMPT_HOOK, &hook, sizeof(hook));
 	return CKR_OK;
 }
 
@@ -595,13 +804,8 @@ pkcs11h_setTokenPromptHook (
 	IN const pkcs11h_hook_token_prompt_t hook,
 	IN void * const global_data
 ) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-	_PKCS11H_ASSERT (hook!=NULL);
-
-	_g_pkcs11h_data->hooks.token_prompt = hook;
-	_g_pkcs11h_data->hooks.token_prompt_data = global_data;
-
+	pkcs11h_setProperty(PKCS11H_PROPERTY_TOKEN_PROMPT_HOOK_DATA, &global_data, sizeof(global_data));
+	pkcs11h_setProperty(PKCS11H_PROPERTY_TOKEN_PROMPT_HOOK, &hook, sizeof(hook));
 	return CKR_OK;
 }
 
@@ -609,11 +813,7 @@ CK_RV
 pkcs11h_setPINCachePeriod (
 	IN const int pin_cache_period
 ) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-
-	_g_pkcs11h_data->pin_cache_period = pin_cache_period;
-
+	pkcs11h_setProperty(PKCS11H_PROPERTY_PIN_CACHE_PERIOD, &pin_cache_period, sizeof(pin_cache_period));
 	return CKR_OK;
 }
 
@@ -621,11 +821,7 @@ CK_RV
 pkcs11h_setMaxLoginRetries (
 	IN const unsigned max_retries
 ) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-
-	_g_pkcs11h_data->max_retries = max_retries;
-
+	pkcs11h_setProperty(PKCS11H_PROPERTY_MAX_LOGIN_RETRIES, &max_retries, sizeof(max_retries));
 	return CKR_OK;
 }
 
@@ -633,11 +829,7 @@ CK_RV
 pkcs11h_setProtectedAuthentication (
 	IN const PKCS11H_BOOL allow_protected_auth
 ) {
-	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
-	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-
-	_g_pkcs11h_data->allow_protected_auth = allow_protected_auth;
-
+	pkcs11h_setProperty(PKCS11H_PROPERTY_ALLOW_PROTECTED_AUTHENTICATION, &allow_protected_auth, sizeof(allow_protected_auth));
 	return CKR_OK;
 }
 
@@ -650,6 +842,375 @@ pkcs11h_addProvider (
 	IN const unsigned slot_event_method,
 	IN const unsigned slot_poll_interval,
 	IN const PKCS11H_BOOL cert_is_private
+) {
+	CK_RV rv;
+
+	if ((rv = pkcs11h_registerProvider(reference)) != CKR_OK) {
+		goto cleanup;
+	}
+	if ((rv = pkcs11h_setProviderProperty(reference, PKCS11H_PROVIDER_PROPERTY_LOCATION, provider_location, strlen(provider_location) + 1)) != CKR_OK) {
+		goto cleanup;
+	}
+	if ((rv = pkcs11h_setProviderProperty(reference, PKCS11H_PROVIDER_PROPERTY_ALLOW_PROTECTED_AUTH, &allow_protected_auth, sizeof(allow_protected_auth))) != CKR_OK) {
+		goto cleanup;
+	}
+	if ((rv = pkcs11h_setProviderProperty(reference, PKCS11H_PROVIDER_PROPERTY_MASK_PRIVATE_MODE, &mask_private_mode, sizeof(mask_private_mode))) != CKR_OK) {
+		goto cleanup;
+	}
+	if ((rv = pkcs11h_setProviderProperty(reference, PKCS11H_PROVIDER_PROPERTY_SLOT_EVENT_METHOD, &slot_event_method, sizeof(slot_event_method))) != CKR_OK) {
+		goto cleanup;
+	}
+	if ((rv = pkcs11h_setProviderProperty(reference, PKCS11H_PROVIDER_PROPERTY_SLOT_POLL_INTERVAL, &slot_poll_interval, sizeof(slot_poll_interval))) != CKR_OK) {
+		goto cleanup;
+	}
+	if ((rv = pkcs11h_setProviderProperty(reference, PKCS11H_PROVIDER_PROPERTY_CERT_IS_PRIVATE, &cert_is_private, sizeof(cert_is_private))) != CKR_OK) {
+		goto cleanup;
+	}
+	if ((rv = pkcs11h_initializeProvider(reference)) != CKR_OK) {
+		goto cleanup;
+	}
+
+cleanup:
+
+	if (rv != CKR_OK) {
+		pkcs11h_removeProvider(reference);
+	}
+
+	return rv;
+}
+
+CK_RV
+pkcs11h_registerProvider (
+	IN const char * const reference
+) {
+	_pkcs11h_provider_t provider = NULL;
+	CK_RV rv = CKR_FUNCTION_FAILED;
+
+	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
+	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG2,
+		"PKCS#11: pkcs11h_registerProvider entry version='%s', reference='%s'",
+		PACKAGE_VERSION,
+		reference
+	);
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG1,
+		"PKCS#11: Register provider '%s'",
+		reference
+	);
+
+	if ((rv = _pkcs11h_mem_malloc ((void *)&provider, sizeof (struct _pkcs11h_provider_s))) != CKR_OK) {
+		goto cleanup;
+	}
+
+	if (strlen(reference) + 1 > sizeof(provider->reference)) {
+		goto cleanup;
+	}
+	strcpy (
+		provider->reference,
+		reference
+	);
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG2,
+		"PKCS#11: pkcs11h_registerProvider Provider '%s'",
+		reference
+	);
+
+#if defined(ENABLE_PKCS11H_THREADING)
+	if ((rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.global)) != CKR_OK) {
+		goto cleanup;
+	}
+#endif
+
+	if (_g_pkcs11h_data->providers == NULL) {
+		_g_pkcs11h_data->providers = provider;
+	}
+	else {
+		_pkcs11h_provider_t last = NULL;
+
+		for (
+			last = _g_pkcs11h_data->providers;
+			last->next != NULL;
+			last = last->next
+		);
+		last->next = provider;
+	}
+
+#if defined(ENABLE_PKCS11H_THREADING)
+	_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.global);
+#endif
+
+	rv = CKR_OK;
+
+cleanup:
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG1,
+		"PKCS#11: Provider '%s' registered rv=%lu-'%s'",
+		reference,
+		rv,
+		pkcs11h_getMessage (rv)
+	);
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG2,
+		"PKCS#11: pkcs11h_registerProvider return rv=%lu-'%s'",
+		rv,
+		pkcs11h_getMessage (rv)
+	);
+
+	return rv;
+}
+
+CK_RV
+pkcs11h_setProviderPropertyByName (
+	IN const char * const reference,
+	IN const char * const property_str,
+	IN const char * const value_str
+) {
+	char value[1024];
+	size_t value_size;
+	unsigned property;
+	CK_RV rv = CKR_FUNCTION_FAILED;
+	const char **s;
+
+	property = 0;
+	for (s = __pkcs11h_provider_preperty_names; *s != NULL && strcmp(property_str, *s); s++) {
+		property++;
+	}
+	if (*s == NULL) {
+		goto cleanup;
+	}
+
+	switch(property) {
+		default:
+			goto cleanup;
+		case PKCS11H_PROVIDER_PROPERTY_LOCATION:
+			value_size = strlen(value_str) + 1;
+			if (value_size > sizeof(value)) {
+				goto cleanup;
+			}
+			strcpy(value, value_str);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_SLOT_EVENT_METHOD:
+		case PKCS11H_PROVIDER_PROPERTY_MASK_PRIVATE_MODE:
+		case PKCS11H_PROVIDER_PROPERTY_SLOT_POLL_INTERVAL:
+			*(unsigned *)value = (unsigned)strtol(value_str, 0, 0);
+			value_size = sizeof(unsigned);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_ALLOW_PROTECTED_AUTH:
+		case PKCS11H_PROVIDER_PROPERTY_CERT_IS_PRIVATE:
+			*(PKCS11H_BOOL *)value = (PKCS11H_BOOL)(strtol(value_str, 0, 0) != 0 ? 1 : 0);
+			value_size = sizeof(PKCS11H_BOOL);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_INIT_ARGS:
+			rv = CKR_ATTRIBUTE_TYPE_INVALID;
+			goto cleanup;
+	}
+
+	rv = pkcs11h_setProviderProperty (
+		reference,
+		property,
+		value,
+		value_size
+	);
+
+cleanup:
+
+	return rv;
+}
+
+static
+CK_RV
+__pkcs11h_providerPropertyAddress(
+	IN _pkcs11h_provider_t provider,
+	IN const unsigned property,
+	OUT void ** value,
+	OUT size_t * value_size,
+	OUT PKCS11H_BOOL *do_strdup
+) {
+	CK_RV rv = CKR_FUNCTION_FAILED;
+
+	*do_strdup = FALSE;
+
+	switch (property) {
+		default:
+			_PKCS11H_DEBUG (
+				PKCS11H_LOG_ERROR,
+				"PKCS#11: Trying to lookup unknown provider property '%d'",
+				property
+			);
+			rv = CKR_ATTRIBUTE_TYPE_INVALID;
+			goto cleanup;
+		case PKCS11H_PROVIDER_PROPERTY_LOCATION:
+			*value = &provider->provider_location;
+			*value_size = sizeof(provider->provider_location);
+			*do_strdup = TRUE;
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_ALLOW_PROTECTED_AUTH:
+			*value = &provider->allow_protected_auth;
+			*value_size = sizeof(provider->allow_protected_auth);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_MASK_PRIVATE_MODE:
+			*value = &provider->mask_private_mode;
+			*value_size = sizeof(provider->mask_private_mode);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_SLOT_EVENT_METHOD:
+			*value = &provider->slot_event_method;
+			*value_size = sizeof(provider->slot_event_method);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_SLOT_POLL_INTERVAL:
+			*value = &provider->slot_poll_interval;
+			*value_size = sizeof(provider->slot_poll_interval);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_CERT_IS_PRIVATE:
+			*value = &provider->cert_is_private;
+			*value_size = sizeof(provider->cert_is_private);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_INIT_ARGS:
+			*value = &provider->init_args;
+			*value_size = sizeof(provider->init_args);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_PROVIDER_DESTRUCT_HOOK:
+			*value = &provider->destruct_hook;
+			*value_size = sizeof(provider->destruct_hook);
+		break;
+		case PKCS11H_PROVIDER_PROPERTY_PROVIDER_DESTRUCT_HOOK_DATA:
+			*value = &provider->destruct_hook_data;
+			*value_size = sizeof(provider->destruct_hook_data);
+		break;
+	}
+	rv = CKR_OK;
+
+cleanup:
+	return rv;
+}
+
+CK_RV
+pkcs11h_setProviderProperty (
+	IN const char * const reference,
+	IN const unsigned property,
+	IN const void * value,
+	IN const size_t value_size
+) {
+	_pkcs11h_provider_t provider = NULL;
+	void *target;
+	size_t size;
+	PKCS11H_BOOL do_strdup;
+	CK_RV rv = CKR_OK;
+
+	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
+	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
+	_PKCS11H_ASSERT (value!=NULL);
+
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG2,
+		"PKCS#11: pkcs11h_setProviderProperty entry reference='%s', property='%d', value=%p, value_size=%ld",
+		reference,
+		property,
+		value,
+		value_size
+	);
+
+	if ((provider = __pkcs11h_get_pkcs11_provider(reference)) == NULL) {
+		rv = CKR_OBJECT_HANDLE_INVALID;
+		goto cleanup;
+	}
+
+	if ((rv = __pkcs11h_providerPropertyAddress(provider, property, &target, &size, &do_strdup)) != CKR_OK) {
+		goto cleanup;
+	}
+
+	if (do_strdup) {
+		const char *str = (const char *)value;
+		if (
+			target != NULL &&
+			(rv = _pkcs11h_mem_free((void *)target)) != CKR_OK
+		) {
+			goto cleanup;
+		}
+		if ((rv = _pkcs11h_mem_strdup(target, str)) != CKR_OK) {
+			goto cleanup;
+		}
+
+		_PKCS11H_DEBUG (
+			PKCS11H_LOG_DEBUG1,
+			"PKCS#11: Setting provider property %s:%s=%s",
+			reference,
+			__pkcs11h_provider_preperty_names[property],
+			str
+		);
+	}
+	else {
+		if (size != value_size) {
+			rv = CKR_DATA_LEN_RANGE;
+			goto cleanup;
+		}
+
+		if (value_size == sizeof(int)) {
+			_PKCS11H_DEBUG (
+				PKCS11H_LOG_DEBUG1,
+				"PKCS#11: Setting provider property %s:%s=0x%x",
+				reference,
+				__pkcs11h_provider_preperty_names[property],
+				*(int *)value
+			);
+		}
+		else if (value_size == sizeof(long)) {
+			_PKCS11H_DEBUG (
+				PKCS11H_LOG_DEBUG1,
+				"PKCS#11: Setting provider property %s:%s=0x%lx",
+				reference,
+				__pkcs11h_provider_preperty_names[property],
+				*(long *)value
+			);
+		}
+		else {
+			_PKCS11H_DEBUG (
+				PKCS11H_LOG_DEBUG1,
+				"PKCS#11: Setting provider property %s:%s=*size*",
+				reference,
+				__pkcs11h_provider_preperty_names[property]
+			);
+		}
+
+		memcpy(target, value, size);
+	}
+	rv = CKR_OK;
+
+	switch (property) {
+		case PKCS11H_PROVIDER_PROPERTY_LOCATION:
+			strncpy (
+				provider->manufacturerID,
+				(
+					strlen (provider->provider_location) < sizeof (provider->manufacturerID) ?
+					provider->provider_location :
+					provider->provider_location+strlen (provider->provider_location)-sizeof (provider->manufacturerID)+1
+				),
+				sizeof (provider->manufacturerID)-1
+			);
+			provider->manufacturerID[sizeof (provider->manufacturerID)-1] = '\x0';
+		break;
+	}
+
+cleanup:
+	_PKCS11H_DEBUG (
+		PKCS11H_LOG_DEBUG2,
+		"PKCS#11: pkcs11h_setProviderProperty return rv=%lu-'%s'",
+		rv,
+		pkcs11h_getMessage (rv)
+	);
+
+	return rv;
+}
+
+CK_RV
+pkcs11h_initializeProvider (
+        IN const char * const reference
 ) {
 #if defined(ENABLE_PKCS11H_DEBUG)
 #if defined(_WIN32)
@@ -664,65 +1225,36 @@ pkcs11h_addProvider (
 
 	_pkcs11h_provider_t provider = NULL;
 	CK_C_GetFunctionList gfl = NULL;
-	CK_C_INITIALIZE_ARGS initargs;
-	CK_C_INITIALIZE_ARGS_PTR pinitargs = NULL;
+	CK_C_INITIALIZE_ARGS init_args;
+	CK_C_INITIALIZE_ARGS_PTR pinit_args = NULL;
 	CK_INFO info;
 	CK_RV rv = CKR_FUNCTION_FAILED;
 
 	_PKCS11H_ASSERT (_g_pkcs11h_data!=NULL);
 	_PKCS11H_ASSERT (_g_pkcs11h_data->initialized);
-	_PKCS11H_ASSERT (provider_location!=NULL);
-	/*_PKCS11H_ASSERT (szSignMode!=NULL); NOT NEEDED*/
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_addProvider entry version='%s', pid=%d, reference='%s', provider_location='%s', allow_protected_auth=%d, mask_private_mode=%08x, cert_is_private=%d",
-		PACKAGE_VERSION,
+		"PKCS#11: pkcs11h_initializeProvider entry pid=%d, reference='%s'",
 		mypid,
-		reference,
-		provider_location,
-		allow_protected_auth ? 1 : 0,
-		mask_private_mode,
-		cert_is_private ? 1 : 0
+		reference
 	);
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG1,
-		"PKCS#11: Adding provider '%s'-'%s'",
-		reference,
-		provider_location
+		"PKCS#11: Initializing provider '%s'",
+		reference
 	);
 
-	if ((rv = _pkcs11h_mem_malloc ((void *)&provider, sizeof (struct _pkcs11h_provider_s))) != CKR_OK) {
+	if ((provider = __pkcs11h_get_pkcs11_provider(reference)) == NULL) {
+		rv = CKR_OBJECT_HANDLE_INVALID;
 		goto cleanup;
 	}
 
-	strncpy (
-		provider->reference,
-		reference,
-		sizeof (provider->reference)-1
-	);
-	provider->reference[sizeof (provider->reference)-1] = '\x0';
-	strncpy (
-		provider->manufacturerID,
-		(
-			strlen (provider_location) < sizeof (provider->manufacturerID) ?
-			provider_location :
-			provider_location+strlen (provider_location)-sizeof (provider->manufacturerID)+1
-		),
-		sizeof (provider->manufacturerID)-1
-	);
-	provider->manufacturerID[sizeof (provider->manufacturerID)-1] = '\x0';
-	provider->allow_protected_auth = allow_protected_auth;
-	provider->mask_private_mode = mask_private_mode;
-	provider->slot_event_method = slot_event_method;
-	provider->slot_poll_interval = slot_poll_interval;
-	provider->cert_is_private = cert_is_private;
-
 #if defined(_WIN32)
-	provider->handle = LoadLibraryA (provider_location);
+	provider->handle = LoadLibraryA (provider->provider_location);
 #else
-	provider->handle = dlopen (provider_location, RTLD_NOW);
+	provider->handle = dlopen (provider->provider_location, RTLD_NOW | RTLD_LOCAL);
 #endif
 
 	if (provider->handle == NULL) {
@@ -758,12 +1290,17 @@ pkcs11h_addProvider (
 		goto cleanup;
 	}
 
-	memset(&initargs, 0, sizeof(initargs));
-	if ((initargs.pReserved = getenv("PKCS11H_INIT_ARGS_RESERVED")) != NULL) {
-		pinitargs = &initargs;
+	if (provider->init_args != NULL) {
+		pinit_args = provider->init_args;
+	}
+	else {
+		memset(&init_args, 0, sizeof(init_args));
+		if ((init_args.pReserved = getenv("PKCS11H_INIT_ARGS_RESERVED")) != NULL) {
+			pinit_args = &init_args;
+		}
 	}
 
-	if ((rv = provider->f->C_Initialize (pinitargs)) != CKR_OK) {
+	if ((rv = provider->f->C_Initialize (pinit_args)) != CKR_OK) {
 		if (rv == CKR_CRYPTOKI_ALREADY_INITIALIZED) {
 			rv = CKR_OK;
 		}
@@ -787,44 +1324,18 @@ pkcs11h_addProvider (
 
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_addProvider Provider '%s' manufacturerID '%s'",
+		"PKCS#11: pkcs11h_initializeProvider Provider '%s' manufacturerID '%s'",
 		reference,
 		provider->manufacturerID
 	);
 
 	provider->enabled = TRUE;
 
-#if defined(ENABLE_PKCS11H_THREADING)
-	if ((rv = _pkcs11h_threading_mutexLock (&_g_pkcs11h_data->mutexes.global)) != CKR_OK) {
-		goto cleanup;
-	}
-#endif
-
-	if (_g_pkcs11h_data->providers == NULL) {
-		_g_pkcs11h_data->providers = provider;
-	}
-	else {
-		_pkcs11h_provider_t last = NULL;
-
-		for (
-			last = _g_pkcs11h_data->providers;
-			last->next != NULL;
-			last = last->next
-		);
-		last->next = provider;
-	}
-
-	provider = NULL;
-
-#if defined(ENABLE_PKCS11H_THREADING)
-	_pkcs11h_threading_mutexRelease (&_g_pkcs11h_data->mutexes.global);
-#endif
-
 	rv = CKR_OK;
 
 cleanup:
 
-	if (provider != NULL) {
+	if (provider != NULL && !provider->enabled) {
 		if (provider->handle != NULL) {
 #if defined(_WIN32)
 			FreeLibrary (provider->handle);
@@ -833,26 +1344,16 @@ cleanup:
 #endif
 			provider->handle = NULL;
 		}
-
-		_pkcs11h_mem_free ((void *)&provider);
-		provider = NULL;
 	}
 
+
 #if defined(ENABLE_PKCS11H_SLOTEVENT)
 	_pkcs11h_slotevent_notify ();
 #endif
 
-	_PKCS11H_DEBUG (
-		PKCS11H_LOG_DEBUG1,
-		"PKCS#11: Provider '%s' added rv=%lu-'%s'",
-		reference,
-		rv,
-		pkcs11h_getMessage (rv)
-	);
-
 	_PKCS11H_DEBUG (
 		PKCS11H_LOG_DEBUG2,
-		"PKCS#11: pkcs11h_addProvider return rv=%lu-'%s'",
+		"PKCS#11: pkcs11h_initializeProvider return rv=%lu-'%s'",
 		rv,
 		pkcs11h_getMessage (rv)
 	);
@@ -913,15 +1414,7 @@ pkcs11h_removeProvider (
 	}
 #endif
 
-	provider = _g_pkcs11h_data->providers;
-	while (
-		provider != NULL &&
-		strcmp (reference, provider->reference)
-	) {
-		provider = provider->next;
-	}
-
-	if (provider != NULL) {
+	if ((provider = __pkcs11h_get_pkcs11_provider(reference)) != NULL) {
 		provider->enabled = FALSE;
 	}
 
@@ -959,6 +1452,11 @@ free1:
 		goto cleanup;
 	}
 
+	if (provider->destruct_hook != NULL) {
+		provider->destruct_hook(provider->destruct_hook_data, reference);
+		provider->destruct_hook = NULL;
+	}
+
 	provider->reference[0] = '\0';
 
 	if (provider->should_finalize) {
@@ -966,6 +1464,10 @@ free1:
 		provider->should_finalize = FALSE;
 	}
 
+	if (provider->provider_location != NULL) {
+		_pkcs11h_mem_free((void *)&provider->provider_location);
+	}
+
 #if defined(ENABLE_PKCS11H_SLOTEVENT)
 	_pkcs11h_slotevent_notify ();
 
@@ -1345,3 +1847,10 @@ __pkcs11h_forkFixup () {
 
 #endif				/* !WIN32 */
 
+static
+_pkcs11h_provider_t
+__pkcs11h_get_pkcs11_provider(const char * const reference) {
+	_pkcs11h_provider_t provider;
+	for (provider = _g_pkcs11h_data->providers;provider != NULL && strcmp (reference, provider->reference); provider = provider->next);
+	return provider;
+}
